#!/usr/bin/env python3
from stepup.core.api import static
from stepup.reprep.api import convert_inkscape_pdf, convert_inkscape_png

static("smile.svg", "glasses.svg")
convert_inkscape_png("glasses.svg", optional=True)
convert_inkscape_pdf("smile.svg", "final.pdf")



#!/usr/bin/env python3
from stepup.core.api import static
from stepup.reprep.api import compile_typst

static(
    "source/",
    "source/document.typ",
    "source/tada.svg",
    "source/aux/",
    "source/aux/productivity.csv",
)
compile_typst("source/document.typ", "out.pdf", inventory=True, keep_deps=True)



#!/usr/bin/env python3
from stepup.core.api import copy, static
from stepup.reprep.api import sanitize_bibtex

static("references.bib", "paper.aux", "bibsane.yaml")
sanitize_bibtex("references.bib", path_aux="paper.aux", path_cfg="bibsane.yaml")
copy("references.bib", "copy.bib")



#!/usr/bin/env python3
from stepup.core.api import static
from stepup.reprep.api import compile_typst

static("template.typ")
compile_typst("template.typ", "alice.pdf", sysinp={"name": "Alic", "age": 29})
compile_typst("template.typ", "bob.pdf", sysinp={"name": "Bob", "age": 31})
compile_typst("template.typ", "charlie.pdf", sysinp={"name": "Charlie", "age": 27})



#!/usr/bin/env python3
from stepup.core.api import static
from stepup.reprep.api import flatten_latex

static("sub/", "sub/article_structured.tex", "sub/part1.tex", "sub/part2.tex")
flatten_latex("sub/article_structured.tex", "sub/article.tex")



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Unit tests for stepup.reprep.latex_log"""

import pytest
from reprep_common import local_file

from stepup.reprep.latex_log import (
    DEFAULT_MESSAGE,
    MESSAGE_SUFFIX,
    LatexSourceStack,
    parse_latex_log,
)

LATEX_LOG1 = r"""
**article
(./article.tex
LaTeX2e <2022-06-01> patch level 5

[17] [18] [19]
<fig1.pdf, id=90, 200pt x 200pt>
File: fig1.pdf Graphic file (type pdf)
<use fig1.pdf>
Package pdftex.def Info: fig1.pdf  used on input line 300.
(pdftex.def)             Requested size:  200pt x 200pt.
[20] [21]
! Undefined control sequence.
l.396         \begin{center}\foo

The control sequence at the end of the top line
of your error message was never \def'ed. If you have
misspelled it (e.g., `\hobx'), type `I' and the correct
spelling (e.g., `I\hbox'). Otherwise just continue,
and I'll forget about whatever was undefined.

<fig2.pdf, id=97, 100pt x 100pt>
File: fig2.pdf Graphic file (type pdf)
<use fig2.pdf>
"""

LATEX_LOG1_MESSAGE = r"""
! Undefined control sequence.
l.396         \begin{center}\foo
"""


def test_parse_latex_log1(tmpdir):
    with local_file(LATEX_LOG1, "article.log", tmpdir):
        error_info = parse_latex_log("article.log")
    assert error_info.program == "LaTeX"
    assert error_info.src == "./article.tex"
    assert (
        error_info.message.strip()
        == (LATEX_LOG1_MESSAGE + MESSAGE_SUFFIX.format(path="article.log")).strip()
    )


LATEX_LOG2 = r"""
not so much
"""


def test_parse_latex_log2(tmpdir):
    with local_file(LATEX_LOG2, "article.log", tmpdir):
        error_info = parse_latex_log("article.log")
    assert error_info.program == "LaTeX"
    assert error_info.src == "(could not detect source file)"
    assert error_info.message == DEFAULT_MESSAGE.format(path="article.log")


LATEX_LOG3 = r"""
LaTeX Font Info:    Font shape `T1/mdput/m/sl' will be
(Font)              scaled to size 11.28003pt on input line 24.
[1

] (./kwart_cirkelboog_e/divergent.inc.tex

! LaTeX Error: Something's wrong--perhaps a missing \item.

See the LaTeX manual or LaTeX Companion for explanation.
Type  H <return>  for immediate help.
 ...

l.2 \item E
           lektrisch veld opgewekt door continue ladingsverdeling: $\vec{E} ...

Try typing  <return>  to proceed.
If that doesn't work, type  X <return>  to quit.

LaTeX Font Info:    Font shape `OML/mdput/b/n' will be
(Font)              scaled to size 11.28003pt on input line 2.
LaTeX Font Info:    Font shape `OML/mdput/b/n' will be
(Font)              scaled to size 7.52002pt on input line 2.
"""


LATEX_LOG3_MESSAGE = r"""
! LaTeX Error: Something's wrong--perhaps a missing \item.

l.2 \item E
           lektrisch veld opgewekt door continue ladingsverdeling: $\vec{E} ...
"""


def test_parse_latex_log3(tmpdir):
    with local_file(LATEX_LOG3, "article.log", tmpdir):
        error_info = parse_latex_log("article.log")
    assert error_info.program == "LaTeX"
    assert error_info.src == "./kwart_cirkelboog_e/divergent.inc.tex"
    assert (
        error_info.message.strip()
        == (LATEX_LOG3_MESSAGE + MESSAGE_SUFFIX.format(path="article.log")).strip()
    )


LATEX_LOG4 = r"""
**foo
] (./review.tex

(/usr/share/texlive/texmf-dist/tex/latex/mathdesign/mdttfont.def
)

Package hyperref Info: Link coloring ON on input line 29.
(./review.out) (./review.out)
\@outlinefile=\write4
\openout4 = `review.out'.

! Missing $ inserted.
<inserted text>
                $
l.116 \end{gather*}

I've inserted a begin-math/end-math symbol since I think
you left one out. Proceed, with fingers crossed.

LaTeX Font Info:    Font shape `T1/mdput/m/n' will be
(Font)              scaled to size 4.70001pt on input line 116.
! Extra }, or forgotten $.
<template> }
            }\savetaglength@ \endtemplate
l.116 \end{gather*}

I've deleted a group-closing symbol because it seems to be
spurious, as in `$x}$'. But perhaps the } is legitimate and
you forgot something else, as in `\hbox{$x}'. In such cases
the way to recover is to insert both the forgotten and the
deleted material, e.g., by typing `I$}'.

! Extra }, or forgotten $.
<template> }}
             \savetaglength@ \endtemplate
l.116 \end{gather*}

I've deleted a group-closing symbol because it seems to be
spurious, as in `$x}$'. But perhaps the } is legitimate and
you forgot something else, as in `\hbox{$x}'. In such cases
the way to recover is to insert both the forgotten and the
deleted material, e.g., by typing `I$}'.

! Missing $ inserted.
<inserted text>
                $
l.116 \end{gather*}

I've inserted a begin-math/end-math symbol since I think
you left one out. Proceed, with fingers crossed.
"""

LATEX_LOG4_MESSAGE = r"""
! Missing $ inserted.
<inserted text>
                $
l.116 \end{gather*}
"""


def test_parse_latex_log4(tmpdir):
    with local_file(LATEX_LOG4, "article.log", tmpdir):
        error_info = parse_latex_log("article.log")
    assert error_info.program == "LaTeX"
    assert error_info.src == "./review.tex"
    assert (
        error_info.message.strip()
        == (LATEX_LOG4_MESSAGE + MESSAGE_SUFFIX.format(path="article.log")).strip()
    )


LATEX_LOG5 = r"""
This is XeTeX, Version 3.141592653-2.6-0.999994 (TeX Live 2022/CVE-2023-32700 patched)
entering extended mode
 restricted \write18 enabled.
 %&-line parsing enabled.
**solutions
(./solutions.tex
LaTeX2e <2022-06-01> patch level 5
L3 programming layer <2022-12-17> (/usr/share/texlive/texmf-dist/tex/latex/base
/article.cls
Document Class: article 2021/10/04 v1.4n Standard LaTeX document class
(/usr/share/texlive/texmf-dist/tex/latex/base/size12.clo
File: size12.clo 2021/10/04 v1.4n Standard LaTeX file (size option)
)
\bibindent=\dimen140
) (../../../../preamble.inc.tex (/usr/share/texlive/texmf-dist/tex/latex/mathde
sign/mathdesign.sty
))

[1

] (./example/solution.inc.tex
LaTeX Font Info:    Font shape `T1/mdput/m/n' will be
(Font)              scaled to size 7.52002pt on input line 5.
LaTeX Font Info:    Font shape `T1/mdput/m/n' will be
(Font)              scaled to size 4.70001pt on input line 5.
)

! LaTeX Error: Something's wrong--perhaps a missing \item.

See the LaTeX manual or LaTeX Companion for explanation.
Type  H <return>  for immediate help.
 ...

l.40 \end{enumerate}

Try typing  <return>  to proceed.
If that doesn't work, type  X <return>  to quit.
"""

LATEX_LOG5_MESSAGE = r"""
! LaTeX Error: Something's wrong--perhaps a missing \item.

l.40 \end{enumerate}
"""


def test_parse_latex_log5(tmpdir):
    with local_file(LATEX_LOG5, "article.log", tmpdir):
        error_info = parse_latex_log("article.log")
    assert error_info.program == "LaTeX"
    assert error_info.src == "./solutions.tex"
    assert (
        error_info.message.strip()
        == (LATEX_LOG5_MESSAGE + MESSAGE_SUFFIX.format(path="article.log")).strip()
    )


def check_source_stack(latex_log, nline, stack, unfinished):
    lines = latex_log.split("\n")[:nline]
    lss = LatexSourceStack()
    for line in lines:
        lss.feed(line + "\n")
    assert lss.stack == stack
    assert lss.unfinished == unfinished
    assert not lss.unmatched


@pytest.mark.parametrize(
    ("nline", "stack", "unfinished"),
    [
        (7, ["./solutions.tex"], None),
        (
            9,
            ["./solutions.tex"],
            "L3 programming layer <2022-12-17> (/usr/share/texlive/texmf-dist/tex/latex/base",
        ),
        (
            10,
            [
                "./solutions.tex",
                "/usr/share/texlive/texmf-dist/tex/latex/base/article.cls",
            ],
            None,
        ),
        (
            16,
            [
                "./solutions.tex",
                "/usr/share/texlive/texmf-dist/tex/latex/base/article.cls",
            ],
            ") (../../../../preamble.inc.tex (/usr/share/texlive/texmf-dist/tex/latex/mathde",
        ),
        (
            17,
            [
                "./solutions.tex",
                "../../../../preamble.inc.tex",
                "/usr/share/texlive/texmf-dist/tex/latex/mathdesign/mathdesign.sty",
            ],
            None,
        ),
        (18, ["./solutions.tex"], None),
        (24, ["./solutions.tex", "./example/solution.inc.tex"], None),
        (30, ["./solutions.tex"], None),
    ],
)
def test_latex_source_stack5(nline, stack, unfinished):
    check_source_stack(LATEX_LOG5, nline, stack, unfinished)


LATEX_LOG6 = r"""
This is XeTeX, Version 3.141592653-2.6-0.999994 (TeX Live 2022/CVE-2023-32700 patched)
entering extended mode
 restricted \write18 enabled.
 %&-line parsing enabled.
**review
(./review.tex

[5] (./weerstand_serie_en_parallel2/method.inc.tex) (./weerstand_serie_en_paral
lel2/answer.inc.tex)
Overfull \hbox (6.20514pt too wide) in paragraph at lines 315--336
\T1/mdput/m/n/12 -  []
 []

(./weerstand_serie_en_parallel2/solution.inc.tex
File: weerstand_serie_en_parallel2//stappen12.pdf Graphic file (type pdf)
<use weerstand_serie_en_parallel2//stappen12.pdf>
[6]
File: weerstand_serie_en_parallel2//stappen34.pdf Graphic file (type pdf)
<use weerstand_serie_en_parallel2//stappen34.pdf>
) (./interactie_eindige_dipolen/stem.inc.tex
File: interactie_eindige_dipolen//paar_dipolen.pdf Graphic file (type pdf)
<use interactie_eindige_dipolen//paar_dipolen.pdf>
) [7] (./interactie_eindige_dipolen/validation.inc.tex)
! Missing $ inserted.
<inserted text>
                $
l.355

I've inserted something that you may have forgotten.
(See the <inserted text> above.)
With luck, this will get me unwedged. But if you
really didn't forget anything, try typing `2' now; then
my insertion and my current dilemma will both disappear.
"""

LATEX_LOG6_MESSAGE = r"""
! Missing $ inserted.
<inserted text>
                $
l.355
"""


def test_parse_latex_log6(tmpdir):
    with local_file(LATEX_LOG6, "article.log", tmpdir):
        error_info = parse_latex_log("article.log")
    assert error_info.program == "LaTeX"
    assert error_info.src == "./review.tex"
    assert (
        error_info.message.strip()
        == (LATEX_LOG6_MESSAGE + MESSAGE_SUFFIX.format(path="article.log")).strip()
    )


LATEX_LOG7 = r"""
(./pdftexcmds.sty
Package: pdftexcmds 2020-06-27 v0.33 Utility functions of pdfTeX for LuaTeX (HO
)
(./infwarerr.sty
Package: infwarerr 2019/12/03 v1.5 Providing info/warning/error messages (HO)
)
)
"""


@pytest.mark.parametrize(
    ("nline", "stack", "unfinished"),
    [
        (2, ["./pdftexcmds.sty"], None),
        (
            3,
            ["./pdftexcmds.sty"],
            "Package: pdftexcmds 2020-06-27 v0.33 Utility functions of pdfTeX for LuaTeX (HO",
        ),
        (4, ["./pdftexcmds.sty"], None),
        (5, ["./pdftexcmds.sty", "./infwarerr.sty"], None),
        (6, ["./pdftexcmds.sty", "./infwarerr.sty"], None),
        (7, ["./pdftexcmds.sty"], None),
        (8, [], None),
    ],
)
def test_latex_source_stack7(nline, stack, unfinished):
    check_source_stack(LATEX_LOG7, nline, stack, unfinished)


LATEX_LOG8 = r"""
(/usr/share/texlive/texmf-dist/tex/generic/pgf/basiclayer/pgfcorequick.code.tex
File: pgfcorequick.code.tex 2021/05/15 v3.1.9a (3.1.9a)
)
"""


@pytest.mark.parametrize(
    ("nline", "stack", "unfinished"),
    [
        (
            3,
            ["/usr/share/texlive/texmf-dist/tex/generic/pgf/basiclayer/pgfcorequick.code.tex"],
            None,
        ),
    ],
)
def test_latex_source_stack8(nline, stack, unfinished):
    check_source_stack(LATEX_LOG8, nline, stack, unfinished)


LATEX_LOG9 = r"""
(./first.tex
Bluh
(./second.tex))
Blah
"""


@pytest.mark.parametrize(
    ("nline", "stack", "unfinished"),
    [
        (2, ["./first.tex"], None),
        (3, ["./first.tex"], None),
        (4, [], None),
        (5, [], None),
    ],
)
def test_latex_source_stack9(nline, stack, unfinished):
    check_source_stack(LATEX_LOG9, nline, stack, unfinished)


LATEX_LOG10 = r"""
This is XeTeX, Version 3.141592653-2.6-0.999995 (TeX Live 2023)
(preloaded format=xelatex 2023.12.26)  14 JAN 2024 10:12
entering extended mode
 restricted \write18 enabled.
 %&-line parsing enabled.
**questions
(./questions.tex
LaTeX2e <2022-11-01> patch level 1
L3 programming layer <2023-02-22>
(/usr/share/texlive/texmf-dist/tex/latex/base/article.cls
Document Class: article 2022/07/02 v1.4n Standard LaTeX document class
(/usr/share/texlive/texmf-dist/tex/latex/base/size12.clo
File: size12.clo 2022/07/02 v1.4n Standard LaTeX file (size option)
))

LaTeX Warning: File `20.png' not found on input line 549.

! Unable to load picture or PDF file '20.png'.
<to be read again>
                   }
l.549     \answergrid

?
! Emergency stop.
<to be read again>
"""

LATEX_LOG10_MESSAGE = r"""
! Unable to load picture or PDF file '20.png'.
<to be read again>
                   }
l.549     \answergrid
"""


def test_parse_latex_log10(tmpdir):
    with local_file(LATEX_LOG10, "questions.log", tmpdir):
        error_info = parse_latex_log("questions.log")
    assert error_info.program == "LaTeX"
    assert error_info.src == "./questions.tex"
    print(error_info.message.strip())
    print("####")
    print((LATEX_LOG10_MESSAGE + MESSAGE_SUFFIX.format(path="questions.log")).strip())
    assert (
        error_info.message.strip()
        == (LATEX_LOG10_MESSAGE + MESSAGE_SUFFIX.format(path="questions.log")).strip()
    )



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Unit tests for stepup.reprep.unplot."""

import json

import numpy as np
from numpy.testing import assert_allclose

from stepup.reprep.unplot import main


def test_bearing_speed_effect(path_tmp):
    # Plot taken from:
    # https://en.m.wikipedia.org/wiki/File:Measured_Bearing_Speed_Effect_data_and_curve.jpg
    path_out = path_tmp / "out.json"
    main(["docs/advanced_topics/unplot/plot.svg", path_out])
    with open(path_out) as fh:
        data = json.load(fh)
    assert data["units"]["mill speed"] == "mpm"
    assert data["units"]["screw movement"] == "micron"
    x = data["curves"]["measured"]["mill speed"]
    y = data["curves"]["measured"]["screw movement"]
    assert abs(x[0] - 199) < 1
    assert abs(y[0] - 165) < 1
    assert abs(x[-1] - 1099) < 1
    assert abs(y[-1] - 382) < 1
    assert len(x) == 10
    assert len(y) == 10


def test_allosteric_modulator(path_tmp):
    # Plot taken from:
    # https://en.m.wikipedia.org/wiki/File:Negative_allosteric_modulator_plot.svg
    path_out = path_tmp / "out.json"
    main(["tests/examples/unplot/plot1.svg", path_out])
    with open(path_out) as fh:
        data = json.load(fh)
    xl = "agonist concentration"
    yl = "response"
    assert data["units"][xl] == "unknown"
    assert data["units"][yl] == "percent"
    curves = data["curves"]
    xi = curves["initial"][xl]
    yi = curves["initial"][yl]
    assert abs(xi[0] - 1e-2) < 1e-5
    assert abs(yi[0] - 1) < 0.1
    assert abs(xi[-1] - 2500) < 20
    assert abs(yi[-1] - 85) < 1
    assert len(xi) == 49
    assert len(yi) == 49
    xs = curves["second"][xl]
    ys = curves["second"][yl]
    assert abs(xs[0] - 1e-2) < 1e-5
    assert abs(ys[0] - 1) < 0.1
    assert abs(xs[-1] - 2500) < 20
    assert abs(ys[-1] - 85) < 1
    assert len(xs) == 33
    assert len(ys) == 33
    xt = curves["third"][xl]
    yt = curves["third"][yl]
    print(xt)
    print(yt)
    assert abs(xt[0] - 1e-2) < 1e-5
    assert abs(yt[0] - 1) < 0.1
    assert abs(xt[-1] - 2500) < 20
    assert abs(yt[-1] - 64) < 1
    assert len(xt) == 13
    assert len(yt) == 13


def test_simple(path_tmp):
    path_out = path_tmp / "out.json"
    main(["tests/examples/unplot/plot2.svg", path_out])
    with open(path_out) as fh:
        data = json.load(fh)
    assert data["units"]["h"] == "1"
    assert data["units"]["v"] == "1"
    x = np.array(data["curves"]["rect"]["h"])
    y = np.array(data["curves"]["rect"]["v"])
    assert x.shape == (4,)
    assert y.shape == (4,)
    assert_allclose(x, [1, 1, 10, 10], atol=1e-3)
    assert_allclose(y, [-1, 1, 1, -1], atol=1e-3)



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Unit tests for stepup.reprep.bibtex_log"""

from reprep_common import local_file

from stepup.reprep.bibtex_log import DEFAULT_MESSAGE, parse_bibtex_log

BIBTEX_BLG1 = """\
This is BibTeX, Version 0.99d (TeX Live 2022/CVE-2023-32700 patched)
Capacity: max_strings=200000, hash_size=200000, hash_prime=170003
The top-level auxiliary file: article.aux
The style file: achemso.bst
Reallocated singl_function (elt_size=4) to 100 items from 50.
Database file #1: acs-article.bib
Database file #2: references.bib
I was expecting a `{' or a `('---line 12 of file references.bib
 :
 : @article{SomeAuthor1999,
(Error may have been on previous line)
I'm skipping whatever remains of this entry
achemso 2022-11-25 v3.13f
You've used 62 entries,
            2538 wiz_defined-function locations,
            1232 strings with 23159 characters,
and the built_in function-call counts, 27532 in all, are:
"""

BIBTEX_BLG1_MESSAGE = """\
I was expecting a `{' or a `('---line 12 of file references.bib
 :
 : @article{SomeAuthor1999,
(Error may have been on previous line)
I'm skipping whatever remains of this entry
"""


def test_parse_bibtex_log1(tmpdir):
    with local_file(BIBTEX_BLG1, "bibtex.blg", tmpdir):
        error_info = parse_bibtex_log("bibtex.blg")
    assert error_info.program == "BibTeX"
    assert error_info.src == "references.bib"
    assert error_info.message == BIBTEX_BLG1_MESSAGE.strip()


BIBTEX_BLG2 = """\
This is BibTeX, Version 0.99d (TeX Live 2022/CVE-2023-32700 patched)
There are not obvious problems in the log file.
"""


def test_parse_bibtex_log2(tmpdir):
    with local_file(BIBTEX_BLG2, "bibtex.blg", tmpdir):
        error_info = parse_bibtex_log("bibtex.blg")
    assert error_info.program == "BibTeX"
    assert error_info.src == "(could not detect source file)"
    assert error_info.message == DEFAULT_MESSAGE.format(path="bibtex.blg")


BIBTEX_BLG3 = r"""\
This is BibTeX, Version 0.99d (TeX Live 2023)
Capacity: max_strings=200000, hash_size=200000, hash_prime=170003
The top-level auxiliary file: reply.aux
I found no \bibstyle command---while reading file reply.aux
You've used 15 entries,
            0 wiz_defined-function locations,
            114 strings with 812 characters,
and the built_in function-call counts, 0 in all, are:
"""


def test_parse_bibtex_log3(tmpdir):
    with local_file(BIBTEX_BLG3, "bibtex.blg", tmpdir):
        error_info = parse_bibtex_log("bibtex.blg")
    assert error_info.program == "BibTeX"
    assert error_info.src == "reply.aux"
    assert error_info.message == r"I found no \bibstyle command---while reading file reply.aux"


BIBTEX_BLG4 = r"""\
This is BibTeX, Version 0.99d (TeX Live 2023)
Capacity: max_strings=200000, hash_size=200000, hash_prime=170003
The top-level auxiliary file: article.aux
The style file: achemso.bst
White space in argument---line 78 of file article.aux
 : \citation{lagauche_thermodynamic_2017
 :                                       pigeon_revisiting_2022}
I'm skipping whatever remains of this command
Reallocated singl_function (elt_size=4) to 100 items from 50.
"""


BLG_ERROR4 = """\
White space in argument---line 78 of file article.aux
 : \\citation{lagauche_thermodynamic_2017
 :                                       pigeon_revisiting_2022}
I'm skipping whatever remains of this command"""


def test_parse_bibtex_log4(tmpdir):
    with local_file(BIBTEX_BLG4, "bibtex.blg", tmpdir):
        error_info = parse_bibtex_log("bibtex.blg")
    print(error_info)
    assert error_info.program == "BibTeX"
    assert error_info.src == "article.aux"
    assert error_info.message == BLG_ERROR4



#!/usr/bin/env python3
from path import Path

from stepup.core.api import static
from stepup.reprep.api import compile_typst

static("template.typ", "persons.json")
compile_typst("template.typ", "persons.pdf", sysinp={"json": Path("persons.json")})



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
"""Sanitize BibTeX files."""

import argparse
import enum
import os
import re
import shutil
import sys
import tempfile
from collections.abc import Collection

import attrs
import bibtexparser
import cattrs
import yaml
from path import Path
from pyiso4.ltwa import Abbreviate

from stepup.core.hash import compute_file_digest

__all__ = ("BibsaneConfig", "main")


@enum.unique
class DuplicatePolicy(enum.Enum):
    """The three policies for duplicate entries."""

    FAIL = "fail"
    MERGE = "merge"
    IGNORE = "ignore"


@enum.unique
class FieldPolicy(enum.Enum):
    """The two policies for bibliography fields."""

    MUST = "must"
    MAY = "may"


@attrs.define
class BibsaneConfig:
    """The configuration object controling BibSane behavior.

    Note that the settings default to the most permissive and least invasive ones.
    We recommend the opposite settings, but you have to switch knowingly in the config file.
    """

    root: Path = attrs.field()
    """The parent directory of the configuration file."""

    drop_entry_types: list[str] = attrs.field(default=attrs.Factory(list))
    """The entry types to drop from the BibTeX database."""

    normalize_doi: bool = attrs.field(default=False)
    """Set to `True` to normalize the DOIs in the entries."""

    duplicate_id: DuplicatePolicy = attrs.field(default=DuplicatePolicy.IGNORE)
    """The policy for duplicate BibTeX IDs: fail, merge or ignore."""

    duplicate_doi: DuplicatePolicy = attrs.field(default=DuplicatePolicy.IGNORE)
    """The policy for duplicate DOIs: fail, merge or ignore."""

    preambles_allowed: bool = attrs.field(default=True)
    """Set to `False` to disallow @preamble entries in the BibTeX database."""

    normalize_whitespace: bool = attrs.field(default=False)
    """Set to `True` to normalize the whitespace in the field values."""

    normalize_names: bool = attrs.field(default=False)
    """Set to `True` to normalize the author and editor names.

    (This currently broken.)
    """

    fix_page_double_hyphen: bool = attrs.field(default=False)
    """Set to `True` to fix the page ranges for which no double hyphen is used."""

    abbreviate_journals: bool = attrs.field(default=True)

    custom_abbreviations: dict[str, str] = attrs.field(factory=dict)
    """Custom journal abbreviations.

    By default, pyiso4 is used to abbreviate journal names.
    The custom abbreviations can override those provided by pyiso4.
    """

    sort: bool = attrs.field(default=False)
    """Set to `True` to sort the entries by year and first author.

    The sort key is `{year}{first author lowercase last name}`.
    """

    citation_policies: dict[str, dict[str, FieldPolicy]] = attrs.field(default=attrs.Factory(dict))
    """The field policies (must or may) for each entry type."""

    @classmethod
    def from_file(cls, fn_yaml: str):
        """Instantiate a configuration from a YAML config file."""
        if fn_yaml is None:
            config = cls(os.getcwd())
        else:
            with open(fn_yaml) as f:
                data = yaml.safe_load(f)
                data.setdefault("root", os.path.dirname(fn_yaml))
                config = cattrs.structure(data, cls)
        return config


RETURN_CODE_SUCCESS = 0
RETURN_CODE_CHANGED = 1
RETURN_CODE_BROKEN = 2

# List of citations to ignore, which are added by some LaTeX templates,
# but which are not correctly parsed by python-bibtexparser.
# Related issue: https://github.com/sciunto-org/python-bibtexparser/issues/384
IGNORED_CITATIONS = {"REVTEX41Control", "achemso-control"}


def main(argv: list[str] | None = None):
    """Main program."""
    fn_bib, fn_aux, verbose, path_out, config = parse_args(argv)
    return process_aux(fn_bib, fn_aux, verbose, path_out, config)


def parse_args(argv: list[str] | None = None) -> tuple[list[str], bool, BibsaneConfig]:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser("rr-bibsane")
    parser.add_argument("bib", help="The BibTeX file to check and clean up.")
    parser.add_argument(
        "-a",
        "--aux",
        help="The LaTeX aux file of your document. If given, unused entries will be dropped.",
    )
    parser.add_argument("-q", "--quiet", default=False, action="store_true")
    parser.add_argument("-c", "--config", help="An optional configuration file")
    parser.add_argument(
        "--out",
        help="Output path for the cleaned up BibTeX file. "
        "If not given, rr-bibsane will be overwrite the given file with the cleaned up version "
        "(if any changes were needed). "
        "This will cause StepUp to drain the scheduler so you can inspect the changes and rebuild.",
    )
    args = parser.parse_args(argv)
    config = BibsaneConfig.from_file(args.config)
    return args.bib, args.aux, not args.quiet, args.out, config


def process_aux(
    fn_bib: str, fn_aux: str | None, verbose: bool, path_out: Path | None, config: BibsaneConfig
) -> int:
    """Main program."""
    # Load the bib file.
    if verbose:
        print("📂 Loading", fn_bib)
    entries, valid_duplicates = collect_entries(fn_bib, config)
    if verbose:
        print(f"   Found {len(entries)} BibTeX entries")
    retcode = RETURN_CODE_CHANGED if valid_duplicates else RETURN_CODE_BROKEN

    # Load the aux file.
    if fn_aux is not None:
        if not fn_aux.endswith(".aux"):
            if verbose:
                print("Please, give an aux file as command-line argument, got:", fn_aux)
            return RETURN_CODE_BROKEN
        if verbose:
            print("📂 Loading", fn_aux)
        citations = parse_aux(fn_aux)
        if verbose:
            print(f"   Found {len(citations)} citations")
        citations = set(citations)
        if verbose:
            print(f"   Found {len(citations)} unique citations")
        if len(citations) == 0:
            if verbose:
                print("   ❓ Ignoring aux file because there are no citations.")
        else:
            # Drop unused and check for missing
            if verbose:
                print("🔨 Checking unused and missing citations")
            entries, bibdata_complete = drop_check_citations(
                entries, citations, config.drop_entry_types
            )
            if not bibdata_complete:
                retcode = RETURN_CODE_BROKEN
            if verbose:
                print(f"   Found {len(entries)} used BibTeX entries")

    # Clean entries
    if len(config.citation_policies) > 0:
        if verbose:
            print("🔨 Validating citation policies")
        entries, valid_fields = clean_entries(entries, config.citation_policies)
        if not valid_fields:
            retcode = RETURN_CODE_BROKEN

    # Clean up things that should never be there, not optional
    if verbose:
        print("🔨 Fixing bad practices")
    entries = fix_bad_practices(entries)

    # Check for potential problems that cannot be fixed automatically, not optional.
    if verbose:
        print("🔨 Checking for potential mistakes in BibTeX keys")
    if potential_mistakes(entries):
        retcode = RETURN_CODE_BROKEN

    # Normalize the DOIs (lowercase and remove prefix)
    if config.normalize_doi:
        if verbose:
            print("🔨 Normalizing dois")
        entries, valid_dois = normalize_doi(entries)
        if not valid_dois:
            retcode = RETURN_CODE_BROKEN

    # Remove newlines
    if config.normalize_whitespace:
        if verbose:
            print("🔨 Normalizing whitespace")
        entries = normalize_whitespace(entries)

    # Normalize author and editor names
    if config.normalize_names:
        if verbose:
            print("🔨 Normalizing author and editor names")
        entries = normalize_names(entries)

    # Fix page double hyphen
    if config.fix_page_double_hyphen:
        if verbose:
            print("🔨 Fixing double hyphen in page ranges")
        entries = fix_page_double_hyphen(entries)

    # Abbreviate journal names
    if config.abbreviate_journals:
        if verbose:
            print("🔨 Abbreviating journal names")
        entries = abbreviate_journal_iso(entries, config.custom_abbreviations)

    # Merge entries
    if config.duplicate_id == DuplicatePolicy.MERGE:
        if verbose:
            print("🔨 Merging references by BibTeX ID")
        entries, merge_conflict = merge_entries(entries, "ID")
        if merge_conflict:
            retcode = RETURN_CODE_BROKEN
        if verbose:
            print(f"   Reduced to {len(entries)} BibTeX entries by merging duplicate BibTeX IDs")
    if config.duplicate_doi == DuplicatePolicy.MERGE:
        if verbose:
            print("🔨 Merging references by DOI")
        entries, merge_conflict = merge_entries(entries, "doi")
        if merge_conflict:
            retcode = RETURN_CODE_BROKEN
        if verbose:
            print(f"   Reduced to {len(entries)} BibTeX entries by merging duplicate DOIs")

    # Sort entries
    if config.sort:
        if verbose:
            print("🔨 Sorting by Year + First author")
        entries = sort_entries(entries)

    # Overwrite if needed.
    fn_out = fn_bib if path_out is None else path_out
    retcode = write_output(entries, fn_out, retcode, verbose)
    if path_out is not None and retcode == RETURN_CODE_CHANGED:
        retcode = RETURN_CODE_SUCCESS
    return retcode


def parse_aux(fn_aux: str) -> tuple[list[str], list[str]]:
    """Parse the relevant parts of a LaTeX aux file."""
    citations = []
    with open(fn_aux) as f:
        for line in f:
            parse_aux_line("citation", line, citations)
    # Filter out bogus citations
    return [citation for citation in citations if citation not in IGNORED_CITATIONS]


def parse_aux_line(prefix: str, line: str, words: list[str]):
    """Parse a (simple) line from a LaTeX aux file."""
    if line.startswith(rf"\{prefix}{{"):
        assert line.endswith("}\n")
        assert line.count("{") == 1
        assert line.count("}") == 1
        words.extend(line[line.find("{") + 1 : -2].split(","))


def collect_entries(fn_bib: str, config: BibsaneConfig) -> tuple[list[dict[str, str]], bool]:
    """Collect entries from multiple BibTeX files."""
    # Collect stuff
    seen_ids = set()
    seen_dois = set()
    entries = []
    valid = True
    bibtex_parser = bibtexparser.bparser.BibTexParser(
        homogenize_fields=True,
        ignore_nonstandard_types=False,
    )
    with open(fn_bib) as f:
        db_in = bibtexparser.load(f, bibtex_parser)
    if len(db_in.preambles) > 0 and not config.preambles_allowed:
        print("   🤖 @preamble is not allowed")
        valid = False
    for entry in db_in.entries:
        if entry["ID"] in seen_ids and config.duplicate_id == DuplicatePolicy.FAIL:
            print(f"  ‼️ Duplicate BibTeX entry: {entry['ID']}")
            valid = False
        if "doi" in entry:
            if entry["doi"] in seen_dois and config.duplicate_doi == DuplicatePolicy.FAIL:
                print(f"‼  ️ Duplicate DOI: {entry['doi']}")
                valid = False
            seen_dois.add(entry["doi"])
        entries.append(entry)
        seen_ids.add(entry["ID"])
    return entries, valid


def drop_check_citations(
    entries: list[dict[str, str]], citations: Collection[str], drop
) -> tuple[list[dict[str, str]], bool]:
    """Drop unused citations and complain about missing ones."""
    # Check for undefined references
    defined = {entry["ID"] for entry in entries}
    valid = True
    for citation in citations:
        if citation not in defined:
            print("   💀 Missing reference:", citation)
            valid = False

    # Drop unused and irrelevant entries
    result = []
    for entry in entries:
        if entry["ID"] not in citations:
            print("     Dropping unused id:", entry["ID"])
            continue
        if entry["ENTRYTYPE"] in drop:
            print("     Dropping irrelevant entry type:", entry["ENTRYTYPE"])
            continue
        result.append(entry)

    return result, valid


def clean_entries(
    entries: list[dict[str, str]], citation_policies: dict[str, dict[str, FieldPolicy]]
) -> tuple[list[dict[str, str]], bool]:
    """Clean the irrelevant fields in each entry and complain about missing ones."""
    cleaned = []
    valid = True
    for old_entry in entries:
        eid = old_entry.pop("ID")
        etype = old_entry.pop("ENTRYTYPE")
        new_entry = {"ENTRYTYPE": etype, "ID": eid}
        if "bibsane" in old_entry:
            etype = old_entry.pop("bibsane")
            new_entry["bibsane"] = etype
        entry_policy = citation_policies.get(etype)
        if entry_policy is None:
            print(f"   🤔 {eid}: @{etype} is not configured")
            valid = False
            continue
        cleaned.append(new_entry)
        for field, policy in entry_policy.items():
            if policy == FieldPolicy.MUST:
                if field not in old_entry:
                    print(f"   🫥 {eid}: @{etype} missing field {field}")
                    valid = False
                else:
                    new_entry[field] = old_entry.pop(field)
            else:
                assert policy == FieldPolicy.MAY
                if field in old_entry:
                    new_entry[field] = old_entry.pop(field)
        if len(old_entry) > 0:
            for field in old_entry:
                print(f"   💨 {eid}: @{etype} discarding field {field}")
    return cleaned, valid


def fix_bad_practices(entries: list[dict[str, str]]) -> list[dict[str, str]]:
    """Fix unwarranted use of braces."""
    result = []
    for old_record in entries:
        # Strip all braces
        new_record = {
            key: value.replace("{", "").replace("}", "") for (key, value) in old_record.items()
        }
        # Except from the author, editor, note or title
        for field in "author", "editor", "note", "title":
            if field in old_record:
                new_record[field] = old_record[field]
        result.append(new_record)
    return result


def potential_mistakes(entries: list[dict[str, str]]) -> bool:
    """Detect potential mistakes in the BibTeX entry keys."""
    id_case_map = {}
    for entry in entries:
        id_case_map.setdefault(entry["ID"].lower(), []).append(entry["ID"])

    mistakes = False
    for groups in id_case_map.values():
        if len(groups) > 1:
            print("   👻 BibTeX entry keys that only differ by case:", " ".join(groups))
            mistakes = True
    return mistakes


DOI_PROXIES = [
    "https://doi.org/",
    "http://doi.org/",
    "http://dx.doi.org/",
    "https://dx.doi.org/",
    "doi:",
]


def normalize_doi(entries: list[dict[str, str]]) -> tuple[list[dict[str, str]], bool]:
    """Normalize the DOIs in the entries."""
    result = []
    valid = True
    for entry in entries:
        doi = entry.get("doi")
        if doi is None:
            new_entry = entry
        else:
            doi = doi.lower()
            for proxy in DOI_PROXIES:
                if doi.startswith(proxy):
                    doi = doi[len(proxy) :]
                    break
            if doi.count("/") == 0 or not doi.startswith("10."):
                print("   🤕 invalid DOI:", doi)
                valid = False
            new_entry = entry | {"doi": doi}
        result.append(new_entry)
    return result, valid


def normalize_whitespace(entries: list[dict[str, str]]) -> list[dict[str, str]]:
    """Normalize the whitespace inside the field values."""
    return [{key: re.sub(r"\s+", " ", value) for key, value in entry.items()} for entry in entries]


def normalize_names(entries: list[dict[str, str]]) -> list[dict[str, str]]:
    """Normalize the author and editor names."""
    raise NotImplementedError("processing of names is not robust in BibtexParser 1.4.0")
    result = []
    for entry in entries:
        # Warning: bibtexparser modifies entries in place.
        # It does not hurt in this case, but it can otherwise give unexpected results.
        new_entry = entry
        for field in "author", "editor":
            if field in entry:
                splitter = getattr(bibtexparser.customization, field)
                new_entry = splitter(new_entry)
                names = entry[field]
                names = [bibtexparser.latexenc.latex_to_unicode(name) for name in names]
                names = [bibtexparser.latexenc.string_to_latex(name) for name in names]
                entry[field] = " and ".join(names)
        result.append(new_entry)
    return result


def fix_page_double_hyphen(entries: list[dict[str, str]]) -> list[dict[str, str]]:
    """Fix page ranges for which no double hyphen is used."""
    return [
        # Warning: bibtexparser modifies entries in place.
        # It does not hurt in this case, but it can otherwise give unexpected results.
        bibtexparser.customization.page_double_hyphen(entry)
        for entry in entries
    ]


def abbreviate_journal_iso(
    entries: list[dict[str, str]], custom: dict[str, str]
) -> list[dict[str, str]]:
    """Replace journal names by their ISO abbreviation."""
    # Abbreviate journals
    result = []
    abbreviator = Abbreviate.create()
    for entry in entries:
        journal = entry.get("journal")
        new_entry = entry
        if journal is not None and "." not in journal:
            abbrev = custom.get(journal)
            if abbrev is None:
                abbrev = abbreviator(journal, remove_part=True)
            abbrev = custom.get(abbrev, abbrev)
            new_entry = new_entry | {"journal": abbrev}
        result.append(new_entry)
    return result


def merge_entries(entries: list[dict[str, str]], field: str) -> tuple[list[dict[str, str]], bool]:
    """Merge entries who have the same value for the given field. (case-insensitive)"""
    lookup = {}
    missing_key = []
    merge_conflict = False
    for entry in entries:
        identifier = entry.get(field)
        if identifier is None:
            print(f"   👽 Cannot merge entry without {field}:", entry["ID"])
            missing_key.append(entry)
        else:
            other = lookup.setdefault(identifier, {})
            for key, value in entry.items():
                if key not in other:
                    other[key] = value
                elif other[key] != value:
                    print(f"   😭 Same {field}={identifier}, different {key}:", value, other[key])
                    merge_conflict = True
    return list(lookup.values()) + missing_key, merge_conflict


def sort_entries(entries: list[dict[str, str]]) -> list[dict[str, str]]:
    """Sort the entries in convenient way: by year, then by author."""

    def keyfn(entry):
        # Make a fake entry to avoid in-place modification.
        entry = {"author": entry.get("author", "Aaaa Aaaa"), "year": entry.get("year", "0000")}
        first_author = bibtexparser.customization.author(entry)["author"][0].lower()
        return entry["year"] + first_author

    return sorted(entries, key=keyfn)


def write_output(entries: list[dict[str, str]], fn_out: str, retcode: int, verbose: bool) -> int:
    """Write out the fixed bibtex file, in case it has changed."""
    if retcode == RETURN_CODE_CHANGED:
        # Write out a single BibTeX database.
        db_out = bibtexparser.bibdatabase.BibDatabase()
        db_out.entries = entries
        writer = bibtexparser.bwriter.BibTexWriter()
        writer.order_entries_by = None
        with tempfile.TemporaryDirectory("rr-bibsane") as dn_tmp:
            fn_tmp = os.path.join(dn_tmp, "tmp.bib")
            with open(fn_tmp, "w") as f:
                bibtexparser.dump(db_out, f, writer)
            if os.path.isfile(fn_out):
                old_hash = compute_file_digest(fn_out)
                new_hash = compute_file_digest(fn_tmp)
                if old_hash == new_hash:
                    retcode = RETURN_CODE_SUCCESS
            if retcode == RETURN_CODE_CHANGED:
                print("💾 Please check the new or corrected file:", fn_out)
                shutil.copy(fn_tmp, fn_out)
            elif verbose:
                print("😀 No changes to", fn_out)
    else:
        print(f"💥 Broken bibliography. Not writing: {fn_out}")

    return retcode


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
from stepup.core.api import runsh, static
from stepup.reprep.api import compile_typst

static("../doc/", "../doc/main.typ", "../doc/other.typ")
compile_typst("../doc/main.typ")
runsh(
    "echo 'This sentence is generated by another step.' > ${out}",
    out="generated.typ",
    workdir="../doc",
    optional=True,
)
